#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <langinfo.h>
#include <iconv.h>
#include <dlfcn.h>
#include <dirent.h>
#include <getopt.h>
#include <errno.h>
#include <sys/stat.h>

#define UHU_MENU_CORE_ENGINE
#include "uhumenu.h"
#include "build-tree.h"

#define MENU_SRC_DIR "/usr/lib/menu/SOURCE"
#define MENU_DST_DIR "/usr/lib/menu"
#define PLUGIN_DIR "/usr/lib/libuhumenu"

const char *menu_src_dir = MENU_SRC_DIR;
const char *menu_dst_dir = MENU_DST_DIR;
const char *plugin_dir = PLUGIN_DIR;

char buf[PATH_MAX];

char *wm_only = NULL;
char *wm_skip = NULL;

locale_t locale = 0;

#define PLUGINS_MAX 64
struct plugin {
	char name[FILENAME_MAX + 1];
	int ok;
	int type;
};

struct plugin plugins[PLUGINS_MAX];
int pluginnr = 0;


int comma_substring (const char *haystack, const char *needle)
{
	int retval;
	char *h2 = malloc(strlen(haystack) + 3);
	char *n2 = malloc(strlen(needle) + 3);
	sprintf(h2, ",%s,", haystack);
	sprintf(n2, ",%s,", needle);
	retval = (strstr(h2, n2) != NULL);
	free(n2);
	free(h2);
	return retval;
}

int plugin_sort_fn (const void *a1, const void *a2)
{
	const struct plugin *p1 = a1;
	const struct plugin *p2 = a2;
	if (p1->type != p2->type) return p1->type - p2->type;
	return strcmp(p1->name, p2->name);
}

void plugins_sort ()
{
	qsort(plugins, pluginnr, sizeof(struct plugin), plugin_sort_fn);
}

char *get_name_lang (const struct item *item, const char *lang)
{
	char buf[32];
	char *s;

	sprintf(buf, "Name[%s]", lang);
	s = get_value(item, buf);
	if (s != NULL) return s;
	s = get_value(item, "Name");
	if (s != NULL) return s;
	return "";
}

int gagyisort (const struct item *a1, const struct item *a2)
{
	if (a1->valid != a2->valid) return a2->valid - a1->valid;
	if (a1->issubmenu != a2->issubmenu) return a2->issubmenu - a1->issubmenu;
	if (a1->major != a2->major) return a2->major - a1->major;
	if (a1->minor != a2->minor) return a2->minor - a1->minor;
	return strcoll_l(get_name_lang(a1, "hu"), get_name_lang(a2, "hu"), locale);
}

void qsort_menu (struct menu *menu, int (*cmp)(const struct item *, const struct item *))
{
	int i;
	int (*compare)(const void *, const void *);

	compare = (int (*)(const void *, const void *)) cmp;
	qsort(menu->items, menu->nr_real, sizeof(struct item), compare);
	for (i = 0; i < menu->nr_real; i++) {
		if (menu->items[i].issubmenu) {
			qsort_menu(menu->items[i].submenu, cmp);
		}
	}
}

int prepare_menu_for_wm (struct menu *menu, const char *wm)
{
	int i;
	menu->nr = 0;
	for (i = 0; i < menu->nr_real; i++) {
		menu->items[i].valid = 1;
		if (menu->items[i].issubmenu) {
			if (prepare_menu_for_wm(menu->items[i].submenu, wm) < 0) {
				menu->items[i].valid = 0;
			}
		}
		if (menu->items[i].valid) {
			if (menu->items[i].wmlist != NULL &&
			  !comma_substring(menu->items[i].wmlist, wm)) {
				menu->items[i].valid = 0;
			}
		}
		if (menu->items[i].valid) menu->nr++;
	}
	if (menu->nr == 0) return -1;
	return 0;
}

void list ()
{
	int i;
	for (i = 0; i < pluginnr; i++) {
		if (!plugins[i].ok) continue;
		printf("%-15s (%s)\n",
		  plugins[i].name,
		  plugins[i].type ? "separate config for each language"
		    : "one config for all languages");
	}
}

void help ()
{
	printf(
	  "Usage: uhu-menu [options...]\n"
	  "\n"
	  "  -s, --source-dir       Alternate source dir of menu (default: " MENU_SRC_DIR ")\n"
	  "  -d, --destination-dir  Alternate dst dir of menu (default: " MENU_DST_DIR ")\n"
	  "  -p, --plugin-dir       Alternate plugin dir (default: " PLUGIN_DIR ")\n"
	  "  -l, --list-plugins     List available window manager plugins\n"
	  "      --wm-only          Process only these window managers (comma separated)\n"
	  "      --wm-skip          Skip these window managers (comma separated)\n"
	  "  -h, --help             Guess what :-)\n"
	);
}

int main (int argc, char *argv[])
{
	struct menu *menu;
	void *lib;
	int (*cm) (const struct menu *menu, const char *path);
	int (*cml) (const struct menu *menu, const char *path, const char *lang, iconv_t ic);
	char *s;
	const char *err1, *err2;
	iconv_t ic;
	DIR *d;
	struct dirent *de;
	int c;
	int l;
	int i;
	int opt_list = 0;

	static const struct option long_options[] = {
		{ "source-dir",      1, NULL, 's' },
		{ "src-dir",         1, NULL, 's' },
		{ "destination-dir", 1, NULL, 'd' },
		{ "dest-dir",        1, NULL, 'd' },
		{ "dst-dir",         1, NULL, 'd' },		
		{ "plugin-dir",      1, NULL, 'p' },
		{ "list-plugins",    0, NULL, 'l' },
		{ "wm-only",         1, NULL, 256 },
		{ "wm-skip",         1, NULL, 257 },
		{ "help",            0, NULL, 'h' },
		{ NULL,              0, NULL, 0   }
	};

	setlocale(LC_ALL, "");

	while ((c = getopt_long(argc, argv, "s:d:p:lh", long_options, NULL)) != -1) {
		switch (c) {
		  case 's':
			menu_src_dir = optarg;
			break;
		  case 'd':
			menu_dst_dir = optarg;
			break;
		  case 'p':
			plugin_dir = optarg;
			break;
		  case 'l':
			opt_list = 1;
			break;
		  case 256:
		  case 257:
			if (wm_only != NULL || wm_skip != NULL) {
				fprintf(stderr, "Window manager list already specified\n");
				exit(1);
			}
			if (c == 256) wm_only = optarg;
			else wm_skip = optarg;
			break;
		  case 'h':
			help();
			exit(0);
		  case '?':
		  case ':':
			exit(1);
		}
	}

	d = opendir(plugin_dir);
	if (d == NULL) {
		fprintf(stderr, "Cannot open plugin dir %s: %s\n", plugin_dir, strerror(errno));
		exit(1);
	}
	while ((de = readdir(d)) != NULL) {
		if (!strcmp(de->d_name, "." )) continue;
		if (!strcmp(de->d_name, "..")) continue;
		l = strlen(de->d_name);
		if (l < 15
		  || strncmp(de->d_name, "libuhumenu-", 11)
		  || strcmp(de->d_name + l - 3, ".so")) {
			fprintf(stderr, "Skipping non-plugin %s/%s\n", plugin_dir, de->d_name);
			continue;
		}
		if (pluginnr == PLUGINS_MAX) {
			fprintf(stderr, "Too many plugins found\n");
			exit(1);
		}
		sprintf(buf, "%.*s", l - 14, de->d_name + 11);
		if (wm_only != NULL && !comma_substring(wm_only, buf)) continue;
		if (wm_skip != NULL &&  comma_substring(wm_skip, buf)) continue;
		strcpy(plugins[pluginnr].name, buf);
		plugins[pluginnr].type = 0;
		plugins[pluginnr].ok = 1;
		pluginnr++;
	}
	closedir(d);
	plugins_sort();

	for (i = 0; i < pluginnr; i++) {
		sprintf(buf, "%s/libuhumenu-%s.so", plugin_dir, plugins[i].name);
		lib = dlopen(buf, RTLD_NOW);
		err1 = dlerror();
		if (err1 != NULL) {
			fprintf(stderr, "dlopen %s: %s\n", buf, err1);
			continue;
		}
		dlsym(lib, "create_menu");
		err1 = dlerror();
		dlsym(lib, "create_menu_lang");
		err2 = dlerror();
		if (err1 == NULL && err2 != NULL) plugins[i].type = 0;
		else if (err1 != NULL && err2 == NULL) plugins[i].type = 1;
		else {
			plugins[i].ok = 0;
			fprintf(stderr, "Ignoring faulty plugin %s\n", plugins[i].name);
			dlclose(lib);
			continue;
		}
		dlclose(lib);
	}

	if (opt_list) {
		list();
		exit(0);
	}

	plugins_sort();

	/* innen gny */

	menu = build_menu_tree(menu_src_dir);

	locale = newlocale(LC_ALL_MASK, "hu_HU", NULL);
	s = nl_langinfo_l(CODESET, locale);
	ic = iconv_open(s, "UTF8");

	locale = newlocale(LC_ALL_MASK, "hu_HU.utf8", NULL);

	for (i = 0; i < pluginnr; i++) {
		if (!plugins[i].ok) continue;
		sprintf(buf, "%s/libuhumenu-%s.so", plugin_dir, plugins[i].name);
		lib = dlopen(buf, RTLD_NOW);

		prepare_menu_for_wm(menu, plugins[i].name);
		qsort_menu(menu, gagyisort);

		cm = NULL; cml = NULL; /* make gcc happy */
		if (plugins[i].type == 0) {
			cm = dlsym(lib, "create_menu");
		} else {
			cml = dlsym(lib, "create_menu_lang");
		}
		err1 = dlerror();
		if (err1 != NULL) {
			fprintf(stderr, "Ouch!\n");
			exit(0);
		}

		sprintf(buf, "%s/%s", menu_dst_dir, plugins[i].name);
		mkdir(buf, 0777);
		if (plugins[i].type == 0) {
			(*cm)(menu, buf);
		} else {
			(*cml)(menu, buf, "hu", ic);
		}

		dlclose(lib);
	}

	return 0;
}
